/* ############################# MPC License ############################## */
/* # Wed Nov 19 15:19:19 CET 2008                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # IDDN.FR.001.230040.000.S.P.2007.000.10000                            # */
/* # This file is part of the MPC Runtime.                                # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL "http://www.cecill.info".                              # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - PERACHE Marc marc.perache@cea.fr                                 # */
/* #                                                                      # */
/* ######################################################################## */
#include "sctk_low_level_comm.h"
#include "sctk_debug.h"
#include "sctk_thread.h"
#include "sctk_rpc.h"
#include "sctk_infiniband.h"
#include "sctk_mpcrun_client.h"

#ifdef MPC_USE_INFINIBAND

#include "sctk_rpc.h"
#include <infiniband/verbs.h>
#include "sctk_alloc.h"
#include "sctk_iso_alloc.h"
#include "sctk_net_tools.h"

#define SCTK_MAX_RAIL_NUMBER 4
#define SCTK_MULTIRAIL_THRESHOLD (4*1024)
#define IB_TX_DEPTH 10
#define IB_RX_DEPTH 10
#define IB_MAX_SG_SQ 2
#define IB_MAX_SG_RQ 2
#define IB_MAX_INLINE 128
#define IB_RMDA_DEPTH 2

typedef enum
{
  ibv_entry_used, ibv_entry_perm, ibv_entry_free
} entry_status_t;

typedef struct
{
  void *ptr;
  size_t size;
  struct ibv_mr *mr;
  entry_status_t status;
} ibv_soft_mmu_entry_t;

static int num_devices;

#define MAX_HOST_SIZE 255
static char ibv_hostname[MAX_HOST_SIZE];

#define IBV_ADM_PORT 1

typedef struct
{
  struct ibv_context *context;
  struct ibv_pd *protection_domain;
  struct ibv_port_attr port_attr;
  struct ibv_device_attr device_attr;
  struct ibv_qp **qp;
  struct ibv_cq **of_cq;
  struct ibv_cq **if_cq;
} ibv_device_t;

static ibv_device_t *device_list;

typedef struct
{
  uint16_t rank;
  int device_nb;

  uint16_t lid[SCTK_MAX_RAIL_NUMBER];
  uint32_t *qpn;
  uint32_t psn[SCTK_MAX_RAIL_NUMBER];
  uint32_t rkey[SCTK_MAX_RAIL_NUMBER];
  uint32_t lkey[SCTK_MAX_RAIL_NUMBER];

} ibv_ident_data_t;

static void
ibv_print_ident (ibv_ident_data_t * ident)
{
  int i;
  sctk_debug ("Rank %d", ident->rank);
  sctk_debug ("device_nb %d", ident->device_nb);
  for (i = 0; i < ident->device_nb; i++)
    {
      int j;
      sctk_debug ("device %d rkey %d", i, ident->rkey[i]);
      sctk_debug ("device %d lkey %d", i, ident->lkey[i]);
      sctk_debug ("device %d lid %d", i, ident->lid[i]);
      sctk_debug ("device %d psn %d", i, ident->psn[i]);
      for (j = 0; j < sctk_process_number; j++)
	{
	  sctk_debug ("device %d qpn[%d] %d", i, j,
		      ident->qpn[i + j * SCTK_MAX_RAIL_NUMBER]);
	}
    }
}

static ibv_ident_data_t **peer_list;

/******************************************************************/
/* SOFT MMU                                                       */
/******************************************************************/

static int ibv_soft_mmu_entry_nb = 0;
static ibv_soft_mmu_entry_t **ibv_soft_mmu = NULL;
static sctk_spinlock_t ibv_soft_mmu_lock = 0;

static void
ibv_init_soft_mmu ()
{
  int i;
  ibv_soft_mmu = sctk_malloc (num_devices * sizeof (ibv_soft_mmu_entry_t *));
  for (i = 0; i < num_devices; i++)
    {
      int j;
      ibv_soft_mmu[i] =
	sctk_malloc (device_list[i].device_attr.max_mr *
		     sizeof (ibv_soft_mmu_entry_t));
      for (j = 0; j < device_list[i].device_attr.max_mr; j++)
	{
	  ibv_soft_mmu[i][j].status = ibv_entry_free;
	}
    }

}

static ibv_soft_mmu_entry_t *
ibv_register_soft_mmu_intern (void *ptr, size_t size, int device_number,
			      entry_status_t status)
{
  int i;
  sctk_spinlock_lock (&ibv_soft_mmu_lock);
  for (i = 0; i < device_list[device_number].device_attr.max_mr; i++)
    {
      if (ibv_soft_mmu[device_number][i].status == ibv_entry_free)
	{
	  ibv_soft_mmu[device_number][i].ptr = ptr;
	  ibv_soft_mmu[device_number][i].size = size;
	  ibv_soft_mmu[device_number][i].status = status;
	  ibv_soft_mmu[device_number][i].mr =
	    ibv_reg_mr (device_list[device_number].protection_domain, ptr,
			size,
			IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_LOCAL_WRITE);
	  sctk_nodebug ("Register %p-%p lkey %d rkey %d", ptr,
			(char *) ptr + size,
			ibv_soft_mmu[device_number][i].mr->lkey,
			ibv_soft_mmu[device_number][i].mr->rkey);
	  if (ibv_soft_mmu[device_number][i].mr == NULL)
	    {
	      sctk_error ("Cannot register adm MR.");
	    }
	  ibv_soft_mmu_entry_nb++;
	  sctk_spinlock_unlock (&ibv_soft_mmu_lock);
	  return &(ibv_soft_mmu[device_number][i]);
	}
    }
  sctk_spinlock_unlock (&ibv_soft_mmu_lock);
  return NULL;
}

static ibv_soft_mmu_entry_t *
ibv_register_soft_mmu (void *ptr, size_t size, int device_number)
{
  return ibv_register_soft_mmu_intern (ptr, size, device_number,
				       ibv_entry_used);
}

static ibv_soft_mmu_entry_t *
ibv_register_perm_soft_mmu (void *ptr, size_t size, int device_number)
{
  return ibv_register_soft_mmu_intern (ptr, size, device_number,
				       ibv_entry_perm);
}

static int
ibv_unregister_soft_mmu (void *ptr, size_t size, int device_number)
{
  int i;
  sctk_spinlock_lock (&ibv_soft_mmu_lock);
  for (i = 0; i < device_list[device_number].device_attr.max_mr; i++)
    {
      if (ibv_soft_mmu[device_number][i].ptr == ptr &&
	  ibv_soft_mmu[device_number][i].size == size &&
	  ibv_soft_mmu[device_number][i].status == ibv_entry_used)
	{
	  ibv_dereg_mr (ibv_soft_mmu[device_number][i].mr);
	  ibv_soft_mmu_entry_nb--;
	  ibv_soft_mmu[device_number][i].status = ibv_entry_free;
	  sctk_spinlock_unlock (&ibv_soft_mmu_lock);
	  return 0;
	}
    }
  sctk_spinlock_unlock (&ibv_soft_mmu_lock);
  return 1;
}

/**************************************************************************/
/* RDMA DRIVER                                                            */
/**************************************************************************/
#define SCTK_PENDING_RDMA_NUMBER 10

typedef struct
{
  volatile int ack_number;
  volatile char ack[SCTK_MAX_RAIL_NUMBER];
} sctk_net_rdma_ack_t;

typedef struct
{
  int used;
  struct ibv_sge list;
  struct ibv_send_wr wr;
  struct ibv_send_wr *bad_wr;
  sctk_net_rdma_ack_t *local_ack;
  sctk_net_rdma_ack_t dist_ack;
} sctk_net_rdma_send_t;

typedef struct
{
  sctk_spinlock_t lock;
  sctk_net_rdma_send_t headers[SCTK_PENDING_RDMA_NUMBER];
} sctk_net_rdma_rail_t;

static sctk_net_rdma_rail_t *sctk_rdma_rails;

static void
sctk_net_poll_rdma (int rail)
{
  struct ibv_wc wc[SCTK_PENDING_RDMA_NUMBER];
  int ne;
  int i;
  int j;
  static sctk_spinlock_t lock;

  sctk_spinlock_lock (&lock);
  for (j = 0; j < sctk_process_number; j++)
    {
      if (j != sctk_process_rank)
	{
	  ne =
	    ibv_poll_cq (device_list[rail].of_cq[j], SCTK_PENDING_RDMA_NUMBER,
			 wc);
	  assume (ne >= 0);
	  for (i = 0; i < ne; i++)
	    {
	      sctk_net_rdma_send_t *header;
	      header = &(sctk_rdma_rails[rail].headers[wc[i].wr_id]);
	      sctk_nodebug
		("Status %d on message %d id %d rail %d to %d local addr %p-%p remote %p-%p",
		 wc[i].status, i, wc[i].wr_id, rail, j, header->list.addr,
		 (char *) header->list.addr + header->list.length,
		 (char *) header->wr.wr.rdma.remote_addr,
		 (char *) header->wr.wr.rdma.remote_addr +
		 header->list.length);
	      if (wc[i].status != IBV_WC_SUCCESS)
		{
		  sctk_error ("ERROR Vendor: %d", wc[i].vendor_err);
		}
	      assume (wc[i].status == IBV_WC_SUCCESS);

	      if (sctk_rdma_rails[rail].headers[wc[i].wr_id].local_ack !=
		  NULL)
		{
		  sctk_rdma_rails[rail].headers[wc[i].wr_id].local_ack->
		    ack[rail] = 1;
		}
	      sctk_rdma_rails[rail].headers[wc[i].wr_id].used = 0;
	      sctk_nodebug ("Ack message nb %d on %d", wc[i].wr_id, rail);
	    }
	}
    }
  sctk_spinlock_unlock (&lock);
}

static sctk_net_rdma_send_t *
sctk_net_get_rdma_header (int rail)
{
  sctk_net_rdma_send_t *tmp = NULL;
  int i;
  sctk_spinlock_lock (&(sctk_rdma_rails[rail].lock));
  do
    {
      for (i = 0; i < SCTK_PENDING_RDMA_NUMBER; i++)
	{
	  if (sctk_rdma_rails[rail].headers[i].used == 0)
	    {
	      tmp = &(sctk_rdma_rails[rail].headers[i]);
	      tmp->wr.wr_id = i;
	      tmp->wr.next = NULL;
	      break;
	    }
	}
      sctk_net_poll_rdma (rail);
    }
  while (tmp == NULL);

  tmp->used = 1;
  sctk_spinlock_unlock (&(sctk_rdma_rails[rail].lock));
  return tmp;
}

static sctk_net_rdma_send_t *
sctk_net_rdma_ib_write (int rail, void *local_addr, void *dist_addr,
			size_t size, int process,
			sctk_net_rdma_ack_t * local_ack)
{
  sctk_net_rdma_send_t *header;
  int rc;

  ibv_ident_data_t *dest;
  ibv_ident_data_t *loc;

  dest = peer_list[process];
  loc = peer_list[sctk_process_rank];

  header = sctk_net_get_rdma_header (rail);
  header->local_ack = local_ack;


  header->list.addr = (long) local_addr;
  header->list.length = size;
  header->list.lkey = loc->lkey[rail];

  header->wr.sg_list = &(header->list);
  header->wr.num_sge = 1;
  header->wr.opcode = IBV_WR_RDMA_WRITE;
  header->wr.send_flags = IBV_SEND_SIGNALED;
  header->wr.wr.rdma.remote_addr = (long) dist_addr;
  header->wr.wr.rdma.rkey = dest->rkey[rail];
  header->bad_wr = NULL;

  sctk_nodebug ("Remote key %d Local key %d", header->wr.wr.rdma.rkey,
		header->list.lkey);
  sctk_nodebug ("SEND id %d rail %d ptr local %p-%p dist %p-%p",
		header->wr.wr_id, rail, local_addr,
		(char *) local_addr + size, dist_addr,
		(char *) dist_addr + size);
  rc =
    ibv_post_send (device_list[rail].qp[process], &(header->wr),
		   &(header->bad_wr));
  assume (rc == 0);
  return header;
}

static void
sctk_net_rdma_ib_set_ack_single (int rail, int process, void *dist_ack)
{
  sctk_net_rdma_send_t *header;
  int rc;

  ibv_ident_data_t *dest;
  ibv_ident_data_t *loc;

  dest = peer_list[process];
  loc = peer_list[sctk_process_rank];

  header = sctk_net_get_rdma_header (rail);
  header->local_ack = NULL;

  header->dist_ack.ack_number = 1;
  header->dist_ack.ack[rail] = 1;


  header->list.addr = (long) &(header->dist_ack);
  header->list.length = sizeof (sctk_net_rdma_ack_t);
  header->list.lkey = loc->lkey[rail];

  header->wr.sg_list = &(header->list);
  header->wr.num_sge = 1;
  header->wr.opcode = IBV_WR_RDMA_WRITE;
  header->wr.send_flags = IBV_SEND_SIGNALED;
  header->wr.wr.rdma.remote_addr = (long) dist_ack;
  header->wr.wr.rdma.rkey = dest->rkey[rail];
  header->bad_wr = NULL;

  sctk_nodebug ("SEND ACK id %d rail %d ack local %p-%p dist %p-%p",
		header->wr.wr_id, rail, (char *) header->list.addr,
		(char *) header->list.addr + header->list.length,
		(char *) header->wr.wr.rdma.remote_addr,
		(char *) header->wr.wr.rdma.remote_addr +
		header->list.length);
  rc =
    ibv_post_send (device_list[rail].qp[process], &(header->wr),
		   &(header->bad_wr));
  assume (rc == 0);
}

static int
sctk_net_rdma_select_rail ()
{
#warning "should be optimize"
  return 0;
}

static void sctk_net_rdma_wait (sctk_net_rdma_ack_t * ack);
static int
sctk_net_rdma_write (void *local_addr, void *dist_addr, size_t size,
		     int process, sctk_net_rdma_ack_t * local_ack,
		     sctk_net_rdma_ack_t * dist_ack)
{
#warning "Have to check if addrs are registered"

  if (local_ack != NULL)
    {
      memset ((char *) (local_ack->ack), 0, SCTK_MAX_RAIL_NUMBER);
    }

  sctk_nodebug ("PREPARE WRITE");

  if ((size <= SCTK_MULTIRAIL_THRESHOLD) || (num_devices <= 1))
    {
      /*Single rail approach */
      sctk_net_rdma_send_t *header;
      int rail;

      /*select a rail */
      rail = sctk_net_rdma_select_rail ();

      if (local_ack != NULL)
	{
	  local_ack->ack_number = 1;
	}

      header =
	sctk_net_rdma_ib_write (rail, local_addr, dist_addr, size, process,
				local_ack);

      if (dist_ack != NULL)
	{
	  sctk_net_rdma_ib_set_ack_single (rail, process, dist_ack);
	}

    }
  else
    {
      /*Multirail approach */
      not_implemented ();
    }

  sctk_nodebug ("WRITE DONE");
  return 0;
}
static int
sctk_net_rdma_read (void *local_addr, void *dist_addr, size_t size,
		    int process, sctk_net_rdma_ack_t * ack)
{
  ibv_ident_data_t *dest;
#warning "Have to check if addrs are registered"

  dest = peer_list[process];

  if ((size <= SCTK_MULTIRAIL_THRESHOLD) || (num_devices <= 1))
    {
      /*Single rail approach */
      not_implemented ();
    }
  else
    {
      /*Multirail approach */
      not_implemented ();
    }

  return 0;

}

typedef struct
{
  sctk_net_rdma_ack_t *ack;
  int done;
} sctk_net_rdma_wait_t;

static void
sctk_net_rdma_wait_poll_func (sctk_net_rdma_wait_t * arg)
{
  int i;
  int ack_done = 0;

  if (arg->ack->ack_number <= 0)
    {
      arg->done = 0;
      return;
    }

  for (i = 0; i < num_devices; i++)
    {
      sctk_net_poll_rdma (i);
    }

  for (i = 0; i < SCTK_MAX_RAIL_NUMBER; i++)
    {
      if (arg->ack->ack[i] == 1)
	{
	  ack_done++;
	}
    }
  if (ack_done == arg->ack->ack_number)
    {
      arg->done = 1;
    }
}

static void
sctk_net_rdma_wait (sctk_net_rdma_ack_t * ack)
{
  sctk_net_rdma_wait_t wait_arg;
  int i;
  wait_arg.ack = ack;
  wait_arg.done = 0;
  sctk_thread_wait_for_value_and_poll (&(wait_arg.done), 1,
				       (void (*)(void *))
				       sctk_net_rdma_wait_poll_func,
				       &wait_arg);

  /* WARNING have to reinit the sctk_net_rdma_ack_t structure */
  memset ((char *) (ack->ack), 0, SCTK_MAX_RAIL_NUMBER);
}

static int
sctk_net_rdma_test (sctk_net_rdma_ack_t * ack)
{
  sctk_net_rdma_wait_t wait_arg;
  wait_arg.ack = ack;
  wait_arg.done = 0;

  sctk_net_rdma_wait_poll_func (&wait_arg);
  return wait_arg.done;
}

/******************************************************************/
/* RPC                                                            */
/******************************************************************/
#include "sctk_rpc_rdma.h"

/******************************************************************/
/* COLLECTIVES                                                    */
/******************************************************************/
#define collective_op_steps 32
#define collective_buffer_size (8*1024)
#include "sctk_collectives_rdma.h"

/******************************************************************/
/* PT 2 PT                                                        */
/******************************************************************/
#define SCTK_NET_MESSAGE_NUMBER 128
#define SCTK_NET_MESSAGE_BUFFER (16*1024*SCTK_NET_MESSAGE_NUMBER)
#define SCTK_NET_MESSAGE_SPLIT_THRESHOLD (32*1024)
#include "sctk_pt2pt_rdma.h"

/******************************************************************/
/* DSM                                                            */
/******************************************************************/
#include "sctk_dsm_rdma.h"

/******************************************************************/
/* INIT/END                                                       */
/******************************************************************/
#include "sctk_init_rdma.h"

void
sctk_net_init_driver_infiniband (int *argc, char ***argv)
{
  sctk_net_init_driver_init_modules ("IB");

  sctk_mpcrun_barrier ();
}

#define PRINT_ATTR(name_s,name) \
  sctk_nodebug("\t- %s %lu",name_s,(unsigned long)device_list[i].device_attr.name);

void
sctk_net_preinit_driver_infiniband (void)
{
  int i;
  int j;
  /*
     Do not touch the following variable: infiniband issue if dev_list is global
   */
  struct ibv_device **dev_list;
  ibv_ident_data_t *ibv_ident_data;

  sctk_mpcrun_client_init ();

  gethostname (ibv_hostname, MAX_HOST_SIZE);

  sctk_nodebug ("Init IB driver on %s %d nodes", ibv_hostname,
		sctk_process_number);

  ibv_ident_data = sctk_malloc (sizeof (ibv_ident_data_t) +
				SCTK_MAX_RAIL_NUMBER * sctk_process_number *
				sizeof (uint32_t));
  ibv_ident_data->qpn = ((char *) ibv_ident_data) + sizeof (ibv_ident_data_t);

  dev_list = ibv_get_device_list (&num_devices);
  if (!dev_list)
    {
      sctk_error ("Infiniband: No device found!!!");
    }
  assume (SCTK_MAX_RAIL_NUMBER >= num_devices);
  for (i = 0; i < num_devices; i++)
    {
      sctk_nodebug ("Found device %d %s", i,
		    ibv_get_device_name (dev_list[i]));
    }

  device_list = sctk_malloc (num_devices * sizeof (ibv_device_t));
  for (i = 0; i < num_devices; i++)
    {
      device_list[i].context = ibv_open_device (dev_list[i]);
      sctk_nodebug ("ibv_open_device");
      if (device_list[i].context == NULL)
	{
	  sctk_error ("Cannot get IB context.");
	}
      if (ibv_query_device
	  (device_list[i].context, &device_list[i].device_attr) != 0)
	{
	  sctk_error ("Unable to get device attr.");
	}
      sctk_nodebug ("DEVICE %d:", i);
      sctk_nodebug ("\t- %s %s", "fw_ver", device_list[i].device_attr.fw_ver);
      PRINT_ATTR ("max_mr", max_mr);
    }
  for (i = 0; i < num_devices; i++)
    {
      device_list[i].protection_domain =
	ibv_alloc_pd (device_list[i].context);
      sctk_nodebug ("ibv_alloc_pd");
      if (device_list[i].protection_domain == NULL)
	{
	  sctk_error ("Cannot get IB protection domain.");
	}
    }

  ibv_init_soft_mmu ();
  /* WARNING Traiter le cas multirail */
  for (i = 0; i < num_devices; i++)
    {
      void *ptr;
      size_t size;
      ibv_soft_mmu_entry_t *entry;

      ptr = sctk_isoalloc_buffer;
      size = SCTK_MAX_ISO_ALLOC_SIZE;
      entry = ibv_register_perm_soft_mmu (ptr, size, i);
      ibv_ident_data->rkey[i] = entry->mr->rkey;
      ibv_ident_data->lkey[i] = entry->mr->lkey;
    }
  sctk_rdma_rails =
    sctk_iso_malloc (SCTK_MAX_RAIL_NUMBER * sizeof (sctk_net_rdma_rail_t));



  for (i = 0; i < num_devices; i++)
    {
      device_list[i].of_cq =
	malloc (sctk_process_number * sizeof (struct ibv_cq *));
      memset (device_list[i].of_cq, 0,
	      sctk_process_number * sizeof (struct ibv_cq *));

      device_list[i].if_cq =
	malloc (sctk_process_number * sizeof (struct ibv_cq *));
      memset (device_list[i].if_cq, 0,
	      sctk_process_number * sizeof (struct ibv_cq *));

      device_list[i].qp =
	malloc (sctk_process_number * sizeof (struct ibv_qp *));
      memset (device_list[i].qp, 0,
	      sctk_process_number * sizeof (struct ibv_qp *));
    }

  for (j = 0; j < sctk_process_number; j++)
    {
      if (j != sctk_process_rank)
	{
	  for (i = 0; i < num_devices; i++)
	    {
	      device_list[i].of_cq[j] =
		ibv_create_cq (device_list[i].context, IB_TX_DEPTH, NULL,
			       NULL, 0);
	      assume (device_list[i].of_cq[j]);
	      device_list[i].if_cq[j] =
		ibv_create_cq (device_list[i].context, IB_RX_DEPTH, NULL,
			       NULL, 0);
	      assume (device_list[i].if_cq[j]);
	      sctk_nodebug ("ibv_create_cq x2");
	      {
		struct ibv_qp_init_attr qp_init_attr;
		qp_init_attr.send_cq = device_list[i].of_cq[j];
		qp_init_attr.recv_cq = device_list[i].if_cq[j];
		qp_init_attr.cap.max_send_wr = IB_TX_DEPTH;
		qp_init_attr.cap.max_recv_wr = IB_RX_DEPTH;
		qp_init_attr.cap.max_send_sge = IB_MAX_SG_SQ;
		qp_init_attr.cap.max_recv_sge = IB_MAX_SG_RQ;
		qp_init_attr.cap.max_inline_data = IB_MAX_INLINE;
		qp_init_attr.qp_type = IBV_QPT_RC;

		device_list[i].qp[j] =
		  ibv_create_qp (device_list[i].protection_domain,
				 &qp_init_attr);
		assume (device_list[i].qp[j]);
		sctk_nodebug ("ibv_create_qp");
	      }

	      {
		struct ibv_qp_attr qp_attr;
		qp_attr.qp_state = IBV_QPS_INIT;
		qp_attr.pkey_index = 0;
		qp_attr.port_num = IBV_ADM_PORT;
		qp_attr.qp_access_flags = IBV_ACCESS_REMOTE_WRITE;

		assume (ibv_modify_qp (device_list[i].qp[j], &qp_attr,
				       IBV_QP_STATE |
				       IBV_QP_PKEY_INDEX |
				       IBV_QP_PORT | IBV_QP_ACCESS_FLAGS) ==
			0);
		sctk_nodebug ("ibv_modify_qp 1");
	      }
	      ibv_ident_data->qpn[i + j * SCTK_MAX_RAIL_NUMBER] =
		device_list[i].qp[j]->qp_num;
	      sctk_nodebug ("CREATE QP %d->%d qp_num %d", sctk_process_rank,
			    j, device_list[i].qp[j]->qp_num);
	    }
	}
    }
  for (i = 0; i < num_devices; i++)
    {
      if (ibv_query_port
	  (device_list[i].context, IBV_ADM_PORT, &device_list[i].port_attr))
	{
	  sctk_error ("Couldn't get port attr");
	}
      sctk_nodebug ("ibv_query_port");
      sctk_nodebug ("LID %d", device_list[i].port_attr.lid);
    }

  ibv_ident_data->rank = sctk_process_rank;
  ibv_ident_data->device_nb = num_devices;

  srand48 (getpid () * time (NULL));

  for (i = 0; i < num_devices; i++)
    {
      ibv_ident_data->lid[i] = device_list[i].port_attr.lid;
      ibv_ident_data->psn[i] = lrand48 () & 0xffffff;	/*a corriger */
    }
/*   ibv_print_ident (ibv_ident_data); */

  assume (sctk_mpcrun_client
	  (MPC_SERVER_REGISTER, NULL, 0, ibv_ident_data,
	   sizeof (ibv_ident_data_t) +
	   SCTK_MAX_RAIL_NUMBER * sctk_process_number * sizeof (uint32_t)) ==
	  0);

  peer_list = sctk_malloc (sctk_process_number * sizeof (ibv_ident_data_t *));
  for (i = 0; i < sctk_process_number; i++)
    {
      peer_list[i] = sctk_malloc (sizeof (ibv_ident_data_t) +
				  SCTK_MAX_RAIL_NUMBER * sctk_process_number *
				  sizeof (uint32_t));
      memset (peer_list[i], 0,
	      sizeof (ibv_ident_data_t) +
	      SCTK_MAX_RAIL_NUMBER * sctk_process_number * sizeof (uint32_t));
    }

  /*Implicit barrier */
  for (i = 0; i < sctk_process_number; i++)
    {
      while (sctk_mpcrun_client
	     (MPC_SERVER_GET, peer_list[i], sizeof (ibv_ident_data_t) +
	      SCTK_MAX_RAIL_NUMBER * sctk_process_number * sizeof (uint32_t),
	      &i, sizeof (int)) != 0)
	{
	  sched_yield ();
	}
      peer_list[i]->qpn = ((char *) peer_list[i]) + sizeof (ibv_ident_data_t);
/*       ibv_print_ident (peer_list[i]); */
    }

  for (j = 0; j < sctk_process_number; j++)
    {
      if (j != sctk_process_rank)
	{
	  for (i = 0; i < num_devices; i++)
	    {
	      int offset;

	      offset = sctk_process_rank;
	      sctk_nodebug ("OFFSET %d", offset);

	      assume (device_list[i].qp[j] != NULL);

	      struct ibv_qp_attr attr;
	      {
		/*IB connection QP step RTR */

		attr.qp_state = IBV_QPS_RTR;
		attr.path_mtu = IBV_MTU_1024;
		attr.dest_qp_num =
		  peer_list[j]->qpn[i + offset * SCTK_MAX_RAIL_NUMBER];
		sctk_nodebug
		  ("CONNECT i = %d %d to %d dest_qp_num %d defined %d", i,
		   sctk_process_rank, j, attr.dest_qp_num,
		   device_list[i].qp[j]->qp_num);
		assume (attr.dest_qp_num != 0);

		attr.rq_psn = peer_list[j]->psn[i];
		sctk_nodebug ("PSN to %d device %d psn %d", j, i,
			      attr.rq_psn);
		attr.max_dest_rd_atomic = IB_RMDA_DEPTH;
		attr.min_rnr_timer = 12;
		attr.ah_attr.is_global = 0;
		attr.ah_attr.dlid = peer_list[j]->lid[i];
		sctk_nodebug ("LID to %d device %d lid %d", j, i,
			      attr.ah_attr.dlid);
		attr.ah_attr.sl = 0;
		attr.ah_attr.src_path_bits = 0;
		attr.ah_attr.port_num = IBV_ADM_PORT;

		assume (ibv_modify_qp (device_list[i].qp[j], &attr,
				       IBV_QP_STATE |
				       IBV_QP_AV |
				       IBV_QP_PATH_MTU |
				       IBV_QP_DEST_QPN |
				       IBV_QP_RQ_PSN |
				       IBV_QP_MAX_DEST_RD_ATOMIC |
				       IBV_QP_MIN_RNR_TIMER) == 0);
		sctk_nodebug ("ibv_modify_qp 2");
	      }

	      {
		/*IB connection QP step RTS */

		attr.qp_state = IBV_QPS_RTS;
		attr.timeout = 14;
		attr.retry_cnt = 7;
		attr.rnr_retry = 7;
		attr.sq_psn = ibv_ident_data->psn[i];
		attr.max_rd_atomic = IB_RMDA_DEPTH;
		assume (ibv_modify_qp (device_list[i].qp[j], &attr,
				       IBV_QP_STATE |
				       IBV_QP_TIMEOUT |
				       IBV_QP_RETRY_CNT |
				       IBV_QP_RNR_RETRY |
				       IBV_QP_SQ_PSN |
				       IBV_QP_MAX_QP_RD_ATOMIC) == 0);
		sctk_nodebug ("ibv_modify_qp 3");
	      }
	    }
	}
      else
	{
	  for (i = 0; i < num_devices; i++)
	    {
	      assume (device_list[i].qp[j] == NULL);
	    }
	}
    }


  sctk_nodebug ("allocations");
  sctk_net_preinit_driver_init_modules ();

  sctk_free (ibv_ident_data);
}

#else
void
sctk_net_init_driver_infiniband (int *argc, char ***argv)
{
  not_available ();
}

void
sctk_net_preinit_driver_infiniband (void)
{
  not_available ();
}
#endif
