/* ############################# MPC License ############################## */
/* # Wed Nov 19 15:19:19 CET 2008                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # IDDN.FR.001.230040.000.S.P.2007.000.10000                            # */
/* # This file is part of the MPC Runtime.                                # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL "http://www.cecill.info".                              # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - PERACHE Marc marc.perache@cea.fr                                 # */
/* #                                                                      # */
/* ######################################################################## */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include "sctk_thread_dbg.h"
#include "sctk_debug.h"
#include "sctk_thread.h"
#include "sctk_spinlock.h"
#include "sctk_ethread.h"
#include "sctk_context.h"

#if defined(SCTK_USE_THREAD_DEBUG)
static sctk_spinlock_t sctk_thread_dbg_lock = 0;

static volatile sctk_thread_data_t *mpc_thread_list = NULL;
static volatile unsigned long sctk_thread_number = 0;
static volatile unsigned long sctk_thread_number_alive = 0;
static volatile int sctk_thread_debug_enable = 0;

static char *
sctk_thread_id (sctk_thread_data_t * item)
{
  static char tmp[4096];
  sprintf (tmp, "%06lu task %04d user thread %04d (LWP %02d)",
	   item->thread_number, item->task_id, item->user_thread,
	   item->virtual_processor);
  return (char *) tmp;
}

void
sctk_thread_debug_printf (const char *format, ...)
{
  if (sctk_thread_debug_enable == 1)
    {
      va_list ap;
      char tmp[4096];
      va_start (ap, format);
      vsprintf (tmp, format, ap);
      va_end (ap);
      sctk_noalloc_fwrite (tmp, 1, strlen (tmp), stdout);
      fflush (stdout);
      fflush (stderr);
    }
}


void
sctk_thread_add (sctk_thread_data_t * item)
{
  sctk_spinlock_lock (&sctk_thread_dbg_lock);
  item->thread_number = sctk_thread_number;
  item->tid = sctk_thread_self ();
  item->status = sctk_thread_running_status;

  if (mpc_thread_list == NULL)
    {
      item->prev = item;
      item->next = item;
      mpc_thread_list = item;
    }
  else
    {
      item->next = mpc_thread_list;
      item->prev = mpc_thread_list->prev;
      item->next->prev = item;
      item->prev->next = item;
    }

  sctk_thread_number++;
  sctk_thread_number_alive++;
  sctk_spinlock_unlock (&sctk_thread_dbg_lock);
  sctk_thread_debug_printf ("[New MPCThread %d %p (LWP %d)]\n",
			    item->thread_number,
			    item->tid, item->virtual_processor);
}

void
sctk_thread_remove (sctk_thread_data_t * item)
{
  item->status = sctk_thread_undef_status;
  sctk_spinlock_lock (&sctk_thread_dbg_lock);
  if (item->next == item)
    {
      mpc_thread_list = NULL;
    }
  else
    {
      if (item == mpc_thread_list)
	{
	  mpc_thread_list = item->next;
	}
      item->next->prev = item->prev;
      item->prev->next = item->next;
    }
  sctk_thread_number_alive--;
  sctk_spinlock_unlock (&sctk_thread_dbg_lock);
  sctk_thread_debug_printf ("[End MPCThread %d (LWP %d)]\n",
			    item->thread_number, item->virtual_processor);
}

void
sctk_thread_enable_debug ()
{
  sctk_thread_debug_enable = 1;
  sctk_thread_debug_printf ("Enable MPC thread debug support\n");
}

void
sctk_thread_disable_debug ()
{
  sctk_thread_debug_printf ("Disable MPC thread debug support\n");
  sctk_thread_debug_enable = 0;
}

static inline char *
sctk_get_status (sctk_thread_status_t s)
{
  char *status;
  switch (s)
    {
    case sctk_thread_running_status:
      status = "Running  ";
      break;
    case sctk_thread_blocked_status:
      status = "Blocked  ";
      break;
    case sctk_thread_sleep_status:
      status = "Sleep    ";
      break;
    case sctk_thread_undef_status:
      status = "Undefined";
      break;
    default:
      status = "Unknown  ";
    }
  return status;
}

void
sctk_thread_list ()
{
  sctk_thread_data_t *tmp;
  int i;
  sctk_thread_debug_enable = 1;
  tmp = (sctk_thread_data_t *) mpc_thread_list;
  sctk_thread_debug_printf ("%d MPC Threads\n", sctk_thread_number_alive);
  for (i = 0; i < sctk_thread_number_alive; i++)
    {
      char *th_id;
      char *th_status;
      tmp = (sctk_thread_data_t *) tmp->next;

      th_id = sctk_thread_id (tmp);
      th_status = sctk_get_status (tmp->status);

      sctk_thread_debug_printf ("[MPCThread %s %s]\n", th_id, th_status);
    }
}

void
sctk_thread_list_task (int j)
{
  sctk_thread_data_t *tmp;
  int i;
  sctk_thread_debug_enable = 1;
  tmp = (sctk_thread_data_t *) mpc_thread_list;
  sctk_thread_debug_printf ("MPC Threads for task %d\n", j);
  for (i = 0; i < sctk_thread_number_alive; i++)
    {
      tmp = (sctk_thread_data_t *) tmp->next;
      if (tmp->task_id == j)
	{
	  char *th_id;
	  char *th_status;
	  th_id = sctk_thread_id (tmp);
	  th_status = sctk_get_status (tmp->status);
	  sctk_thread_debug_printf ("[MPCThread %s %s]\n", th_id, th_status);
	}
    }
}

void
sctk_thread_current ()
{
  sctk_thread_data_t *tmp;
  tmp = (sctk_thread_data_t *) sctk_thread_data_get ();
  if (tmp != NULL)
    {
      if (tmp->task_id >= 0)
	{
	  sctk_thread_debug_printf
	    ("[Current thread is %s (Thread %p)]\n",
	     sctk_thread_id (tmp), tmp->tid);
	}
      else
	{
	  sctk_thread_debug_printf
	    ("[Current thread a MPC internal thread]\n");
	}
    }
  else
    {
      sctk_thread_debug_printf ("[Current thread a idle task]\n");
    }
}


void
sctk_thread_goto (int j)
{
  sctk_thread_data_t *tmp = NULL;
  sctk_thread_data_t *step = NULL;
  int i;

  step = (sctk_thread_data_t *) mpc_thread_list;

  for (i = 0; i < sctk_thread_number_alive; i++)
    {
      step = (sctk_thread_data_t *) step->next;
      if (step->thread_number == j)
	tmp = step;
    }

  if (tmp != NULL)
    {
      if (tmp->status == sctk_thread_running_status)
	{
	  sctk_thread_debug_printf
	    ("Unable to switch to a thread in running mode,\nuse mpc_thread_cur localise this thread on LWPs\n");
	}
      else
	{
	  sctk_thread_debug_printf
	    ("The thread %d is in %s state \nuse mpc_print_stack %d to view current state \nof thread %d\n",
	     j, sctk_get_status (tmp->status), j, j);
	}
    }
  else
    {
      sctk_thread_debug_printf ("Thread ID %d not known.\n", j);
    }
}

static volatile sctk_mctx_t old_ctx;
void
sctk_thread_print_stack_in (int j)
{
  int i;
  sctk_thread_data_t *tmp = NULL;
  sctk_thread_data_t *step = NULL;
  static volatile sctk_mctx_t new_ctx;
  sctk_thread_data_t *tmp_old;
  static volatile int done = 1;

  if (done == 1)
    {
      done = 0;
      tmp_old = (sctk_thread_data_t *) sctk_thread_data_get ();

      step = (sctk_thread_data_t *) mpc_thread_list;
      for (i = 0; i < sctk_thread_number_alive; i++)
	{
	  step = (sctk_thread_data_t *) step->next;
	  if (step->thread_number == j)
	    tmp = step;
	}

      if (tmp != NULL)
	{
	  if (tmp->status == sctk_thread_running_status)
	    {
	      sctk_thread_goto (j);
	    }
	  else
	    {
	      if ((tmp->status != sctk_thread_undef_status)
		  && (sctk_ethread_check_state () == 1))
		{
		  char *th_id;
		  char *th_status;
		  th_id = sctk_thread_id (tmp);
		  th_status = sctk_get_status (tmp->status);
		  sctk_thread_debug_printf
		    ("[Switching to thread %s (Thread %p)] %s\n",
		     th_id, tmp->tid, th_status);
		  new_ctx = (((sctk_ethread_per_thread_t *) tmp->tid)->ctx);
		  tmp->force_stop = 1;
		  sctk_swapcontext ((sctk_mctx_t *) & (old_ctx),
				    (sctk_mctx_t *) & (new_ctx));
		}
	      else
		{
		  sctk_thread_debug_printf
		    ("Thread ID %d not available.\n", j);
		}
	    }
	}
      else
	{
	  sctk_thread_debug_printf ("Thread ID %d not known.\n", j);
	}

      if (tmp_old != NULL)
	{
	  if (tmp_old->task_id >= 0)
	    {
	      char *th_id;
	      char *th_status;
	      th_id = sctk_thread_id (tmp_old);
	      th_status = sctk_thread_id (tmp_old);
	      sctk_thread_debug_printf
		("[Switching to thread %s (Thread %p)] %s\n",
		 th_id, tmp_old->tid, th_status);
	    }
	  else
	    {
	      sctk_thread_debug_printf
		("[Switch to an MPC internal thread]\n");
	    }
	}
      else
	{
	  sctk_thread_debug_printf ("[Switch to an idle task]\n");
	}
      done = 1;
    }
  else
    {
      sctk_thread_debug_printf
	("Unable to switch, last switch is not finished\n");
    }
}

void
sctk_thread_print_stack_task (int j)
{
  int i;
  sctk_thread_data_t *step = NULL;

  step = (sctk_thread_data_t *) mpc_thread_list;
  for (i = 0; i < sctk_thread_number_alive; i++)
    {
      step = (sctk_thread_data_t *) step->next;
      if ((step->thread_number == j) && (step->user_thread == 0))
	{
	  sctk_thread_print_stack_in (i);
	  return;
	}
    }
  sctk_thread_debug_printf ("Task ID %d not known.\n", j);
}

void
sctk_thread_print_stack_out ()
{
  /*  static sctk_mctx_t nold_ctx;
     sctk_swapcontext (&(nold_ctx),(sctk_mctx_t*)&(old_ctx)); */
  sctk_setcontext ((sctk_mctx_t *) & (old_ctx));
  abort ();
}


#endif
